# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

from datetime import datetime
import os
import platform
import uuid
import hashlib
import logging
import random
import uuid

""" Utility functions required for initialization of various objects

"""

log = logging.getLogger(__name__)


class systemUUID:
    __UUID = ''

    def __init__(self):
        if systemUUID.__UUID == '':
            self._set_system_uuid()

    # Using uuid.getnode() in order to get the mac address. If all attempts to obtain the hardware address fail, this function returns a random 48-bit number
    # with its eighth bit set to 1 as recommended in RFC 4122. A check on the 8th bit is done in order to figure out if random MAC Address generation is required or not.
    def _getMacAddrIfaceHash(self):
        mac = uuid.getnode()
        if (mac >> 40)%2:
            log.warn( 'could not find a mac address, an unlikely potential exists for uuid collisions with liota instances on other IoT gateways')
            # generate a 48-bit random integer from this seed
            # always returns the same random integer
            # however in get_uuid below a unique uuid for each resource name will be created
            # this allows us not to have to store any uuid on the persistent storage yet
            # create a unique system uuid
            random.seed(1234567)
            mac = random.randint(0, 281474976710655)
        m = hashlib.md5()
        m.update(str(mac))
        self.macHash = m.hexdigest()
        return self.macHash

    def _set_system_uuid(self):
        # start by creating the liota namespace, this is a globally unique uuid, and exactly the same for any instance of liota
        self.liotaNamespace = uuid.uuid5(uuid.NAMESPACE_URL, 'https://github.com/vmware/liota')
        log.info(str('liota namespace uuid: ' + str(self.liotaNamespace)))
        # we create a system uuid for the physical system on which this instance is running
        # we hash the interface name with the mac address in getMacAddrIfaceHash to avoid collision of
        #     mac addresses across potential different physical interfaces in the IoT space
        systemUUID.__UUID = uuid.uuid5(self.liotaNamespace, self._getMacAddrIfaceHash())
        log.info('system UUID: ' + str(systemUUID.__UUID))


    def _get_system_uuid(self):
        if systemUUID.__UUID == '':
            self.set_system_uuid()
        return systemUUID.__UUID


    def get_uuid(self, resource_name):
        # creating a uuid for a particular resource on this system
        # all resources with the same resource name will have the same uuid, on this system
        # resources created on a different system with a name used on this system will have *different* uuids' as it should be
        tmp = str(uuid.uuid5(self._get_system_uuid(), resource_name))
        log.info('resource: ' + resource_name + '  uuid=' + str(tmp))
        return str(tmp)


def get_linux_version():
    return platform.platform()


def getUTCmillis():
    return long(1000 * ((datetime.utcnow() - datetime(1970, 1, 1)).total_seconds()))


class findLiotaConfigFullPath:
    __fullPathLiotaConfig = ''
    syswide_path = '/etc/liota/conf/'

    def __init__(self):
        if findLiotaConfigFullPath.__fullPathLiotaConfig == '':
            self._find_path()

    def _find_path(self):
        """ a multi-step search for the configuration file.
        1. Current working directory. ./liota.conf.
        2. User's home directory (~user/liota.conf)
        3. A place named by an environment variable (LIOTA_CONF)
        4. A standard system-wide directory (such as /etc/liota/conf/liota.conf)
        # assert: every install will have a default liota.conf in /etc/liota/conf
        """

        for loc in os.curdir, os.path.expanduser("~"), os.environ.get(
                "LIOTA_CONF"), findLiotaConfigFullPath.syswide_path:
            if loc is None:
                continue
            path = os.path.join(loc, 'liota.conf')
            if os.path.exists(path):
                findLiotaConfigFullPath.__fullPathLiotaConfig = path
                break
            else:
                continue
        if findLiotaConfigFullPath.__fullPathLiotaConfig == '':
            print 'liota.conf not found'

    def get_liota_fullpath(self):
        return findLiotaConfigFullPath.__fullPathLiotaConfig



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

from abc import ABCMeta, abstractmethod
import logging
import time

from liota.transformers.metrics import Sample
from liota.utilities.utility import getUTCmillis


log = logging.getLogger(__name__)
class Gateway:
    __metaclass__ = ABCMeta
    """ The base class in order to define the various functions that
        are required in order to define Gateway object

    """
    def __init__(self, make, model, os, identifier, res_name, res_uuid, parent, type, res_kind):
        self.make = make
        self.model = model
        self.os = os
        self.identifier = identifier
        self.res_name = res_name
        self.res_uuid = res_uuid
        self.parent = parent
        self.type = type
        self.res_kind = res_kind
        self.property_key_value_map = {}
        # A single pin can report multiple metrics. Following map associates metrics with pins
        self.pin_to_metrics_map = {}
        self.pins = {}
        self._configure_pins()
        self._initialize_gateway()

    @abstractmethod
    def _configure_pins(self):
        pass

    @abstractmethod
    def _initialize_gateway(self):
        pass

    @abstractmethod
    def _report_data(self, msg_id, statkey, timestamps, values):
        return {
         "type": "add_stats",
         "uuid": self.res_uuid,
         "metric_data": [{
            "statKey": statkey,
            "timestamps": timestamps,
            "data": values
         }],
      }

    @abstractmethod
    def _create_relationship(self, msg_id, parent):
        return {
         "transactionID": msg_id,
         "type": "create_relationship_request",
         "body": {
            "parent": parent.res_uuid,
            "child": self.res_uuid
         }
      }



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import inspect
import json
import logging
import sched, time
import signal
from socket import timeout
import sys
import threading
from time import timezone

from liota.dcc.dcc_base import DataCenterComponent
from helix_protocol import HelixProtocol
from liota.core.metric_handler import Metric
from liota.utilities.utility import getUTCmillis


log = logging.getLogger(__name__)

class Vrops(DataCenterComponent):
    """ The implementation of vROPS cloud provider soultion

    """
    def __init__(self, username, password, con):
        log.info("Logging into DCC")
        self.con = con
        self.username = username
        self.password = password
        self.proto = HelixProtocol(self.con, username, password)
        self.resource_uuid = "null"
        def on_receive_safe(msg):
            try:
                log.debug("Received msg: {0}".format(msg))
                json_msg = json.loads(msg)
                self.proto.on_receive(json.loads(msg))
                log.debug("Processed msg: {0}".format(json_msg["type"]))
                if json_msg["type"] == "connection_verified":
                    log.info("Verified")
                    exit()
            except Exception:
                raise
                log.exception("Error received on connecting to DCC instance. Please verify the credentials and try again.")

        thread = threading.Thread(target=self.con.run)
        self.con.on_receive = on_receive_safe
        thread.daemon = True
        thread.start()
        thread.join()
        log.info("Logged in to DCC successfully")

    def register(self, gw):
        """ Register the objects

        """
        if gw.res_uuid == None:
            vrops_res = self.VropsResource(gw)
            log.info("Creating resource")
            log.info("Resource Name: {0}".format(gw.res_name))
            def on_receive_safe(msg):
                try:
                    log.debug("Received msg: {0}".format(msg))
                    if msg != "":
                       json_msg = json.loads(msg)
                       self.proto.on_receive(json.loads(msg))
                       log.debug("Processed msg: {0}".format(json_msg["type"]))
                       if json_msg["type"] == "create_or_find_resource_response" :
                           if json_msg["body"]["uuid"] != "null":
                               log.info("FOUND RESOURCE: {0}".format(json_msg["body"]["uuid"]))
                               gw.res_uuid = json_msg["body"]["uuid"]
                               time.sleep(5)
                               vrops_res.registered = True
                               exit()
                           else:
                               log.info("Waiting for resource creation")
                               time.sleep(5)
                               self.con.send(self.registration(self.con.next_id(), gw.identifier, gw.res_name, gw.res_kind))
                except:
                    raise
            thread = threading.Thread(target=self.con.run)
            self.con.on_receive = on_receive_safe
            thread.daemon = True
            thread.start()
            self.con.send(self.registration(self.con.next_id(), gw.identifier, gw.res_name, gw.res_kind))
            thread.join()
            log.info("Resource Registered {0}".format(gw.res_name))
            gw.con = self.con
            if gw.parent is not None:
                self.init_relations(gw)
                log.info("Relationship Created")
            return vrops_res

    def connect_soc(self, protocol, url, user_name, password):
        pass

    def subscribe(self):
        pass

    def publish(self, metric):
        timestamps = [t for t, _ in metric.values]
        values = [v for _, v in metric.values]
        message = metric.gw._report_data(self.con.next_id(), metric.details, timestamps, values)
        self.con.send(message)

    def init_relations(self, gw):
      """ This function initializes all relations between gateway and it's children.
          It is called after each object's UUID is received.

          Parameters:
          - obj: The object that has just obtained an UUID

      """
      self.con.send(gw._create_relationship(self.con.next_id(), gw.parent))


    def registration(self, msg_id, res_id, res_name, res_kind):
      return {
         "transactionID": msg_id,
         "type": "create_or_find_resource_request",
         "body": {
            "kind": res_kind,
            "id": res_id,
            "name": res_name
         }
      }

    def properties(self, msg_id, res_uuid, res_kind, timestamp, key, value):
        msg = {
            "transationID": msg_id,
            "type": "add_properties",
            "uuid": res_uuid,
            "body": {
                "kind": res_kind,
                "timestamp": timestamp,
                "property_data": []
            }
        }
        msg["body"]["property_data"].append({"propertyKey": key, "propertyValue": value})
        return msg

    def set_properties(self, key, value, registered_gw):
        log.info("Properties defined for resource {0}".format(registered_gw.resource.res_name))
        self.con.send(self.properties(self.con.next_id(), registered_gw.resource.res_uuid, registered_gw.resource.res_kind, getUTCmillis(), key, value))

    class VropsResource:

        def __init__(self, resource, registered=False):
            self.resource = resource
            self.registered = registered



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

#!/usr/bin/env python
import logging
import socket

from transport_layer_base import TransportLayer


log = logging.getLogger(__name__)

class Socket(TransportLayer):

    def __init__(self, carbon_server, carbon_port):
        self.carbon_server = carbon_server
        self.carbon_port = carbon_port
        TransportLayer.__init__(self)

    def connect_soc(self):
        self.sock = socket.socket()
        log.info("Creating Socket")
        try:
            self.sock.connect((self.carbon_server, self.carbon_port))
            log.info("Socket Created")
        except Exception:
            log.exception("Socket connection cannot be established to Graphite DCC. Please check the firewall rules an try again.")
            self.sock.close()
            self.sock = None

    def send(self, message):
        if self.sock is not None:
            self.sock.sendall(message)




# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

from Queue import Queue, PriorityQueue, Full
import heapq
import inspect
import logging
from threading import Thread, Condition
from time import time as _time

from liota.utilities.utility import getUTCmillis

log = logging.getLogger(__name__)

event_ds = None
collect_queue = None
send_queue = None
event_checker_thread = None
send_thread = None

class EventsPriorityQueue(PriorityQueue):
    def __init__(self):
        PriorityQueue.__init__(self)
        self.first_element_changed = Condition(self.mutex)

    def put_and_notify(self, item, block=True, timeout=None):
        log.info("Adding Event:" + str(item))
        self.not_full.acquire()
        try:
            first_element_before_insertion = None
            if self._qsize() > 0:
                first_element_before_insertion = heapq.nsmallest(1, self.queue)[0]

            if self.maxsize > 0:
                if not block:
                    if self._qsize() == self.maxsize:
                        raise Full
                elif timeout is None:
                    while self._qsize() == self.maxsize:
                        self.not_full.wait()
                elif timeout < 0:
                    raise ValueError("'timeout' must be a non-negative number")
                else:
                    endtime = _time() + timeout
                    while self._qsize() == self.maxsize:
                        remaining = endtime - _time()
                        if remaining <= 0.0:
                            raise Full
                        self.not_full.wait(remaining)
            self._put(item)
            self.unfinished_tasks += 1
            self.not_empty.notify()

            first_element_after_insertion = heapq.nsmallest(1, self.queue)[0]
            if first_element_before_insertion != first_element_after_insertion:
                self.first_element_changed.notify()
        finally:
            self.not_full.release()

    def get_next_element_when_ready(self):
        self.first_element_changed.acquire()
        try:
            isNotReady = True
            while isNotReady:
                if self._qsize() > 0:
                    first_element = heapq.nsmallest(1, self.queue)[0]
                    timeout = (first_element.get_next_run_time() - getUTCmillis()) / 1000.0
                    log.info("Waiting on acquired first_element_changed LOCK for: " + str(timeout))
                    self.first_element_changed.wait(timeout)
                else:
                    self.first_element_changed.wait()
                    first_element = heapq.nsmallest(1, self.queue)[0]
                if (first_element.get_next_run_time() - getUTCmillis()) <= 0:
                    isNotReady = False
                    first_element = self._get()
            return first_element
        finally:
            self.first_element_changed.release()

class EventCheckerThread(Thread):

    def __init__(self):
        Thread.__init__(self)
        self.start()

    def run(self):
        log.info("Started EventCheckerThread")
        global event_ds
        global collect_queue
        while True:
            log.debug("Waiting for event...")
            matric = event_ds.get_next_element_when_ready()
            log.debug("Got event:" + str(matric))
            collect_queue.put(matric)

class SendThread(Thread):
    def __init__(self):
        Thread.__init__(self)
        self.start()

    def run(self):
        log.info("Started SendThread")
        global send_queue
        while True:
            log.info("Waiting to send...")
            matric = send_queue.get()
            log.info("Got item in send_queue:" + str(matric))
            matric.send_data()

class CollectionThread(Thread):
    def __init__(self):
        Thread.__init__(self)
        self.daemon = True
        self.start()

    def run(self):
        global event_ds
        global collect_queue
        global send_queue
        while True:
            matric = collect_queue.get()
            log.info("Collecting stats for matric:" + str(matric))
            try:
                matric.collect()
                matric.set_next_run_time()
                event_ds.put_and_notify(matric)
                if matric.is_ready_to_send():
                    send_queue.put(matric)
            except Exception as e:
                log.error(e)

class CollectionThreadPool:
    def __init__(self, num_threads):
        log.info("Starting " + str(num_threads) + " for collection")
        for _ in range(num_threads): CollectionThread()

is_initialization_done = False

def initialize():
    global is_initialization_done
    if is_initialization_done:
        log.debug("Initialization already done")
        pass
    else:
        log.debug("Initializing.............")
        global event_ds
        if event_ds == None:
            event_ds = EventsPriorityQueue()
        global event_checker_thread
        if event_checker_thread == None:
            event_checker_thread = EventCheckerThread()
        global collect_queue
        if collect_queue == None:
            collect_queue = Queue()
        global send_queue
        if send_queue == None:
            send_queue = Queue()
        global send_thread
        if send_thread == None:
            send_thread = SendThread()
        pool = CollectionThreadPool(20)  # TODO: Make pool size configurable
        is_initialization_done = True

class Metric(object):

        def __init__(self, gw, details, unit, sampling_interval_sec, aggregation_size, sampling_function, data_center_component):
            self.data_center_component = data_center_component
            self.gw = gw
            self.details = details
            self.unit = unit
            self.sampling_interval_sec = sampling_interval_sec
            self.aggregation_size = aggregation_size
            self.current_aggregation_size = 0
            self.sampling_function = sampling_function
            self.values = []

        def __str__(self, *args, **kwargs):
            return str(self.details) + ":" + str(self.next_run_time)

        def __cmp__(self, other):
            if other == None:
                return -1
            if not isinstance(other, Metric):
                return -1
            return cmp(self.next_run_time, other.next_run_time)

        def write_full(self, t, v):
            self.values.append((t, v))

        def write_map_values(self, v):
            self.write_full(getUTCmillis(), v)

        def get_next_run_time(self):
            return self.next_run_time

        def set_next_run_time(self):
            self.next_run_time = self.next_run_time + (self.sampling_interval_sec * 1000)
            log.info("Set next run time to:" + str(self.next_run_time))

        def start_collecting(self):
            # TODO: Add a check to ensure that start_collecting for a metric is called only once by the client code
            initialize()
            global event_ds
            self.next_run_time = getUTCmillis() + (self.sampling_interval_sec * 1000)
            event_ds.put_and_notify(self)

        def is_ready_to_send(self):
            log.debug("self.current_aggregation_size:" + str(self.current_aggregation_size))
            log.debug("self.aggregation_size:" + str(self.aggregation_size))
            return self.current_aggregation_size >= self.aggregation_size

        def collect(self):
            log.debug("Collecting values for the resource {0} ".format(self.details))
            self.args_required = len(inspect.getargspec(self.sampling_function)[0])
            if self.args_required is not 0:
                self.cal_value = self.sampling_function(1)
            else:
                self.cal_value = self.sampling_function()
            log.info("{0} Sample Value: {1}".format(self.details, self.cal_value))
            log.debug("Size of the list {0}".format(len(self.values)))
            self.write_map_values(self.cal_value)
            self.current_aggregation_size = self.current_aggregation_size + 1

        def send_data(self):
            log.info("Publishing values {0} for the resource {1} ".format(self.values, self.details))
            if not self.values:
                # No values measured since last report_data
                return True
            self.data_center_component.publish(self)
            self.values[:] = []
            self.current_aggregation_size = 0




# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import logging

from liota.utilities.utility import getUTCmillis


PROTOCOL_VERSION = "2.7"
log = logging.getLogger(__name__)

class HelixProtocolError(RuntimeError):
   pass
class HelixInitializationError(RuntimeError):
   pass

def require_field(container, field):
   if field not in container:
      raise HelixProtocolError("Missing {} field" % field)
      log.error("Missing {} field {0}", format(field))


class HelixProtocol:
   """ Simple state machine for protocol to vROps adapter.
       States and messages:

       handshake_awaiting
       ===================================================
       <-- connection_request

       handshake_responded
       ===================================================
       --> connection_response
       <-- connection_verified

       handshake_verified
       ===================================================
       for all output resource kinds simultaneously do
         --> create_resource_kind_request
         <-- create_resource_kind_response

       steady
       ===================================================
       for all objects simultaneously do
         while uuid not available do
           --> create_or_find_resource_request
           <-- create_or_find_resource_response
         // Create relationship to gateway
         --> create_relationship_request
         <-- create_relationship_response

         loop forever
           if output
             <-- action
           if input
             --> add_stats

   """

   def __init__(self, con, user, password):
      self.con = con
      self.user = user
      self.password = password
      # `HandshakeAwaitingState.__init__` may be using `self.state`,
      # so initialize it.
      self.state = None
      self.state = HandshakeAwaitingState(None, self)

   def on_receive(self, msg):
      require_field(msg, "type")
      require_field(msg, "body")

      if not self.state.on_receive(msg):
         raise HelixProtocolError("Unexpected Message " + msg["type"]
            + " during " + self.state.name)
         log.error("Unexpected Message {0}", msg["type"])

   def transition(self, state):
      self.state = state


class State:

   def __init__(self, previous, proto=None):
      if proto is None:
         proto = previous.proto
      self.proto = proto
      self.con = proto.con
      self.user = proto.user
      self.password = proto.password
      self.name = "Unknown State"
      if self.proto.state == previous:
         self.proto.transition(self)

   def is_active(self):
      return self == self.proto.state

class HandshakeAwaitingState(State):

   def __init__(self, previous, proto=None):
      State.__init__(self, previous, proto)
      self.name = "HandshakeAwaitingState"

   def on_receive(self, msg):
      log.debug("Received message in HandshakeAwaitingState: {0}".format(msg))
      if msg["type"] == "connection_request":
         HandshakeRequestedState(self, msg)
         return True
      else:
         return False


class HandshakeRequestedState(State):

   def __init__(self, previous, msg, proto=None):
      State.__init__(self, previous, proto)
      self.name = "HandshakeRequestedState"
      require_field(msg, "transactionID")
      self.con.send({
         "type": "connection_response",
         "transactionID": msg["transactionID"],
         "body": {
            "version": PROTOCOL_VERSION,
            "username" : self.user,
            "password": self.password
         }
      })

   def on_receive(self, msg):
      log.debug("Received message in HandshakeRequestedState: {0}".format(msg))
      if msg["type"] == "connection_verified":
         require_field(msg["body"], "result")

         if msg["body"]["result"] == "succeeded":
            HandshakeVerifiedState(self)
            return True
         else:
            raise HelixInitializationError("Handshake Failed")
            log.error("Handshake Failed")
      else:
         return False


class HandshakeVerifiedState(State):

   def __init__(self, previous, proto=None):
      State.__init__(self, previous, proto)
      self.name = "HandshakeVerifiedState"
      log.info("Handshake Verified with DCC")
      SteadyState(self)

   def on_receive(self, msg):
      if msg["type"] == "create_resource_kind_response":
         require_field(msg["body"], "result")
         require_field(msg, "transactionID")

         if msg["body"]["result"] == "succeeded" or msg["body"]["result"] == "exists":
            search = [obj for obj in self.obj_list
                      if obj.extra.get("kind_req") == msg["transactionID"]]
            if search:
               search[0].extra["kind_registered"] = True
               search[0].extra["kind_req"] = None
               self.pending_kinds -= 1
               if self.pending_kinds == 0:
                  SteadyState(self)
            return True
         elif msg["body"]["result"] == "pending":
            search = [obj for obj in self.obj_list
                      if obj.extra.get("kind_req") == msg["transactionID"]]
            if search:
               search[0].extra["kind_req"] = None
            return True
         else:
            raise HelixInitializationError("Resource Kind Creation Failed")
            log.error("Resource Kind Creation Failed")
      return False


class SteadyState(State):

   def __init__(self, previous, proto=None):
      State.__init__(self, previous, proto)
      self.name = "SteadyState"
      self.action_map = {}
      log.info("Entered steady state")

   def trigger_action(self, uuid, value):
      if uuid not in self.action_map:
         # ignore unknown actions
         pass
      else:
         handler = self.action_map[uuid]
         handler.on_change(value)

   def on_receive(self, msg):
    log.info("IN ON_RECEIVE")
    if msg["type"] == "create_or_find_resource_response":
       require_field(msg["body"], "uuid")
       require_field(msg, "transactionID")

       res_uuid = msg["body"]["uuid"]
       return True

    elif msg["type"] == "action":
       require_field(msg["body"], "uuid")
       require_field(msg["body"], "code")

       self.trigger_action(msg["body"]["uuid"], msg["body"]["code"])
       return True

    elif msg["type"] == "create_relationship_response":
       # Not checked right now
       return True

    else:
       return False



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

from linux_metrics import cpu_stat, disk_stat, net_stat, mem_stat
from liota.boards.gateway_de5000 import DellEdge5000
from liota.dcc.graphite_dcc import Graphite
from liota.dcc.vrops import Vrops
from liota.things.ram import RAM
from liota.transports.socket_connection import Socket
from liota.transports.web_socket import WebSocket
import random

# getting values from conf file
config = {}
execfile('sampleProp.conf', config)

# some standard metrics for Linux systems
# agent classes for different IoT gateways
# agent classes for different data center components
# agent classes for different kinds of of devices, 'Things', connected to the gw
# we are showing here how to create a representation for a Thing in vROps but
# using the notion of RAM (because we have no connected devies yet)
# agent classes for different kinds of layer 4/5 connections from agent to DCC
# -------User defined functions for getting the next value for a metric --------
# usage of these shown below in main
# semantics are that on each call the function returns the next available value
# from the device or system associated to the metric.
def read_cpu_procs():
    return cpu_stat.procs_running()

def read_cpu_utilization(sample_duration_sec=1):
    cpu_pcts = cpu_stat.cpu_percents(sample_duration_sec)
    return round((100 - cpu_pcts['idle']), 2)

def read_disk_busy_stats(sample_duration_sec=1):
    return round(disk_stat.disk_busy('sda', sample_duration_sec), 4)

def read_mem_free():
    return round((mem_stat.mem_stats()[3]) / (1048576), 3)

def read_network_bits_recieved():
    return round((net_stat.rx_tx_bits('eth0')[0]) / (8192), 2)

def simulated_device():
    return random.randint(0, 20)

#---------------------------------------------------------------------------
# In this example, we demonstrate how gateway health and som simluated data
# can be directed to two data center components (vrops and graphite) using Liota.
# The program illustrates the ease of use Liota brings to IoT application developers.

if __name__ == '__main__':

    # create a data center object, vROps in this case, using websocket as a transport layer
    # this object encapsulates the formats and protocols neccessary for the agent to interact with the dcc
    # UID/PASS login for now.
    vrops = Vrops(config['vROpsUID'], config['vROpsPass'], WebSocket(url=config['WebSocketUrl']))

    # create a gateway object encapsulating the particulars of a gateway/board
    # argument is the name of this gateway
    gateway = DellEdge5000(config['Gateway1Name'])

    # resister the gateway with the vrops instance
    # this call creates a representation (a Resource) in vrops for this gateway with the name given
    vrops_gateway = vrops.register(gateway)

    if vrops_gateway.registered:
        # these call set properties on the Resource representing the gateway in the vrops instance
        # properties are a key:value store
        # arguments are (key, value)
        for item in config['Gateway1PropList']:
            for key, value in item.items():
                vrops.set_properties(key, value, vrops_gateway)
        # ---------- Create metrics 'on' the Resource in vrops representing the gateway
        # arguments:
        #          local object referring to the Resource in vrops on which the metric should be associated
        #          metric name
        #          unit = An SI Unit (work needed here)
        #          sampling_interval = the interval in seconds between called to the user function to obtain the next value for the metric
        #          report_interfal = the interval between subsequent sends to the data center component. If sample > report values are queued
        #          value = user defined function to obtain the next value from the device associated with this metric
        cpu_utilization = vrops.create_metric(vrops_gateway, "CPU_Utilization", unit=None, sampling_interval_sec=50, aggregation_size=2, sampling_function=read_cpu_utilization)

        # call to start collecting values from the device or system and sending to the data center component
        cpu_utilization.start_collecting()

        cpu_procs = vrops.create_metric(vrops_gateway, "CPU_Process", unit=None, sampling_interval_sec=6, sampling_function=read_cpu_procs)
        cpu_procs.start_collecting()

        disk_busy_stats = vrops.create_metric(vrops_gateway, "Disk_Busy_Stats", unit=None, aggregation_size=6, sampling_function=read_disk_busy_stats)
        disk_busy_stats.start_collecting()

        network_bits_recieved = vrops.create_metric(vrops_gateway, "Network_Bits_Recieved", unit=None, sampling_interval_sec=5, sampling_function=read_network_bits_recieved)
        network_bits_recieved.start_collecting()
    else:
        print "vROPS resource not registered successfully"

    # Here we are showing how to create a device object, registering it in vrops, and setting properties on it
    # Since there are no attached devices are as simulating one by considering RAM as separate from the gateway
    # The agent makes possible many different data models
    # arguments:
    #        device name
    #        Read or Write
    #        another Resource in vrops of which the should be the child of a parent-child relationship among Resources
    ram = RAM(config['Device1Name'], 'Read', gateway)
    vrops_device = vrops.register(ram)

    if vrops_device.registered:
        # note that the location of this 'device' is different from the location of the gateway. It's not really different
        # but just an example of how one might create a device different from the gateway
        for item in config['Device1PropList']:
            for key, value in item.items():
                vrops.set_properties(key, value, vrops_device)
        mem_free = vrops.create_metric(vrops_device, "Memory_Free", unit=None, sampling_interval_sec=10, sampling_function=read_mem_free)
        mem_free.start_collecting()
    else:
        print "vROPS resource not registered successfully"


    # Sending data to an alternate data center component (e.g. data lake for analytics)
    # Graphite is a data center component
    # Socket is the transport which the agent uses to connect to the graphite instance
    graphite = Graphite(Socket(config['GraphiteIP'], config['GraphitePort']))
    graphite_gateway = graphite.register(gateway)
    content_metric = graphite.create_metric(graphite_gateway, config['GraphiteMetric'], unit=None, sampling_interval_sec=10, aggregation_size=2, sampling_function=simulated_device)
    content_metric.start_collecting()




# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import logging
import threading
import time

from liota.dcc.dcc_base import DataCenterComponent

log = logging.getLogger(__name__)

class Graphite(DataCenterComponent):

    def __init__(self, socket_obj):
        self.con = socket_obj
        pass

    def publish(self, metric):
        if self.con is not None:
            for t,v in metric.values:
                message = '%s %s %d\n' % (metric.details , v, t/1000) # Graphite expects time in seconds, not milliseconds. Hence, dividing by 1000
                log.info("Sending message: {0}".format(message))
                self.con.send(message)

    def subscribe(self):
        pass

    def register(self, gw):
        graphite_res = self.GraphiteGateway(gw, True)
        return graphite_res

    class GraphiteGateway:
        def __init__(self, gw, registered=False):
            self.resource = gw
            self.registered = registered




# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import json
import logging
import os
import ssl
import sys
from websocket import create_connection

from transport_layer_base import TransportLayer
log = logging.getLogger(__name__)

class WebSocket(TransportLayer):
    """ WebSocket class implementation

    """
    def __init__(self, url):
        self.url = url
        TransportLayer.__init__(self)

    def connect_soc(self):
        try:
            self.WebSocketConnection(self.url, False)
            log.info("Connection Successful")
        except Exception:
            log.exception("WebSocket exception, please check the WebSocket address and try again.")
            sys.exit(0)


    def WebSocketConnection(self, host, verify_cert=True, CERTDIR="/var/lib/helix-agent/cert"):

      if not verify_cert:
            self.ws = None
            self.ws = create_connection(host, enable_multithread=True,
              sslopt={"cert_reqs": ssl.CERT_NONE})
      else:
         self.ws = None
         for filename in os.listdir(CERTDIR):
            if os.path.isfile(CERTDIR + "/" + filename):
               try:
                  self.ws = create_connection(host, enable_multithread=True,
                     sslopt={"cert_reqs": ssl.CERT_REQUIRED, "ca_certs": CERTDIR + "/" + filename})
                  break
               except ssl.SSLError:
                  pass
         if self.ws is None:
            raise(IOError("Couldn't verify host certificate"))

      self.counter = 0

    def run(self):
        try:
            log.debug("Stream Opened")
            while True:
                msg = self.ws.recv()
                log.debug ("Message received while running {0}".format(msg))
                if msg is "":
                    # TO DO Check if os._exit(0) required everywhere
                    log.error("Stream Closed")
                    os._exit(0)
                    break
                log.debug("RX {0}".format(msg))
                if self.on_receive is not None:
                    self.on_receive(msg)
        except Exception:
            log.exception("Exception on receiving the response from Server, please check the connection and try again.")
            self.close()
            os._exit(0)

    def send(self, msg):
      request_calls = ['request', 'response']
      complete_message = json.dumps(msg)
      log.info("TX Sending message {0}".format(complete_message))
      try:
          self.ws.send(complete_message)
      except:
          # Retry logic only for publishing stats, not for request or response calls
          if all(request not in complete_message for request in request_calls):
              attempts = 1
              while attempts < 4:
                  try:
                      log.info("Exception while sending data, applying retry logic.")
                      self.connect_soc()
                      log.info("Created New Websocket")
                      log.info("TX Sending message {0}".format(complete_message))
                      self.ws.send(complete_message)
                      break
                  except Exception:
                      # Three times retry websocket connection for publishing data
                      log.info("{0} attempt".format(attempts))
                      attempts += 1
                      if attempts == 4:
                          log.exception("Exception while sending data, please check the connection and try again.")
                          self.close()
                          os._exit(0)
          else:
              log.exception("Exception while sending data, please check the connection and try again.")
              self.close()
              os._exit(0)

    def next_id(self):
      self.counter = (self.counter + 1) & 0xffffff
      # Enforce even IDs
      return self.counter * 2

    def close(self):
      if self.ws is not None:
          self.ws.close()
      log.debug("Connection closed, cleanup done")



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

class ObjectConfig(object):
   """ Configuration Struct used for Gateway, physical and virtual Inputs/Outputs.

   """
   def __init__(self, identifier, name, local_identifier=None,
                uuid=None, res_kind=None, actions=None,
                stat_key=None, property_key_value_map=None, extra=None, ri=None, mi=None,
                direction=None, template=None, digital_pin=None,
                analog_pin=None, parent=None):
      # global unique identifier for Gateway
      self.identifier = identifier
      # only unique on this device
      self.local_identifier = local_identifier
      # uuid obtained by object registration from vROps
      self.uuid = uuid
      self.name = name
      self.res_kind = res_kind
      self.stat_key = stat_key
      self.property_key_value_map = property_key_value_map
      self.digital_pin = digital_pin
      self.analog_pin = analog_pin
      self.actions = actions
      self.parent = parent
      # extra data that can be associated with this object
      if extra is None:
         self.extra = {}
      else:
         self.extra = extra
      # response interval
      self.ri = ri
      # measurement interval
      self.mi = mi
      # direction (input/output)
      self.direction = direction
      # can be HelixGateway, HelixDigitalPin or HelixAnalogPin
      self.template = template

      # assert self.stat_key is not None or self.direction == "output" or self.template == "HelixGateway"
      # assert self.stat_key is not None
      assert self.identifier is not None and self.identifier != ""
      assert self.name is not None